"""Flagging utility for a MS"""

from __future__ import annotations

from argparse import ArgumentParser
from dataclasses import dataclass
from pathlib import Path

import astropy.units as u

from jolly_roger.baselines import get_baselines_from_ms
from jolly_roger.hour_angles import make_hour_angles_for_ms
from jolly_roger.logging import logger
from jolly_roger.uvws import uvw_flagger, xyz_to_uvw


@dataclass
class JollyRogerFlagOptions:
    """Specifications of the flagging to carry out"""

    min_scale_deg: float = 0.075
    """Minimum angular scale to project to UVW"""
    min_horizon_limit_deg: float = -3
    """The minimum elevation for the sun projected baselines to be considered for flagging"""
    max_horizon_limit_deg: float = 90
    """The minimum elevation for the sun projected baselines to be considered for flagging"""
    dry_run: bool = False
    """Do not apply the flags"""


def flag(ms_path: Path, flag_options: JollyRogerFlagOptions) -> Path:
    # Trust no one
    logger.debug(f"{flag_options=}")

    ms_path = Path(ms_path)
    logger.info(f"Flagging {ms_path=}")

    baselines = get_baselines_from_ms(ms_path=ms_path)
    hour_angles = make_hour_angles_for_ms(ms_path=ms_path, position="sun")

    uvws = xyz_to_uvw(baselines=baselines, hour_angles=hour_angles)
    ms_path = uvw_flagger(
        computed_uvws=uvws,
        min_horizon_lim=flag_options.min_horizon_limit_deg * u.deg,
        max_horizon_lim=flag_options.max_horizon_limit_deg * u.deg,
        min_sun_scale=flag_options.min_scale_deg * u.deg,
        dry_run=flag_options.dry_run,
    )
    logger.info(f"Finished processing {ms_path=}")

    return ms_path


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Flag a measurement set based on properties of the Sun"
    )
    parser.add_argument("ms_path", type=Path, help="The measurement set to flag")

    parser.add_argument(
        "--min-scale-deg",
        type=float,
        default=0.075,
        help="The minimum scale required for flagging",
    )
    parser.add_argument(
        "--min-horizon-limit-deg",
        type=float,
        default=-3,
        help="The minimum elevation of the centroid of the object (e.g. sun) for uvw flagging to be activated",
    )
    parser.add_argument(
        "--max-horizon-limit-deg",
        type=float,
        default=-3,
        help="The maximum elevation of the centroid of the object (e.g. sun) for uvw flagging to be activated",
    )
    parser.add_argument(
        "--dry-run", action="store_true", help="Do not apply the computed flags"
    )

    return parser


def cli() -> None:
    parser = get_parser()

    args = parser.parse_args()

    flag_options = JollyRogerFlagOptions(
        min_scale_deg=args.min_scale_deg,
        min_horizon_limit_deg=args.min_horizon_limit_deg,
        max_horizon_limit_deg=args.max_horizon_limit_deg,
        dry_run=args.dry_run,
    )

    flag(ms_path=args.ms_path, flag_options=flag_options)


if __name__ == "__main__":
    cli()



"""Routines around plotting"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from astropy.visualization import (
    ImageNormalize,
    LogStretch,
    MinMaxInterval,
    SqrtStretch,
    ZScaleInterval,
    quantity_support,
    time_support,
)

from jolly_roger.uvws import WDelays
from jolly_roger.wrap import calculate_nyquist_zone, symmetric_domain_wrap

if TYPE_CHECKING:
    from jolly_roger.delays import DelayTime
    from jolly_roger.tractor import BaselineData


def plot_baseline_data(
    baseline_data: BaselineData,
    output_dir: Path,
    suffix: str = "",
) -> None:
    with quantity_support(), time_support():
        data_masked = baseline_data.masked_data
        data_xx = data_masked[..., 0]
        data_yy = data_masked[..., -1]
        data_stokesi = (data_xx + data_yy) / 2
        amp_stokesi = np.abs(data_stokesi)

        fig, ax = plt.subplots()
        im = ax.pcolormesh(
            baseline_data.time,
            baseline_data.freq_chan,
            amp_stokesi.T,
        )
        fig.colorbar(im, ax=ax, label="Stokes I Amplitude / Jy")
        ax.set(
            ylabel=f"Frequency / {baseline_data.freq_chan.unit:latex_inline}",
            title=f"Ant {baseline_data.ant_1} - Ant {baseline_data.ant_2}",
        )
        output_path = (
            output_dir
            / f"baseline_data_{baseline_data.ant_1}_{baseline_data.ant_2}{suffix}.png"
        )
        fig.savefig(output_path)


def plot_baseline_comparison_data(
    before_baseline_data: BaselineData,
    after_baseline_data: BaselineData,
    before_delays: DelayTime,
    after_delays: DelayTime,
    output_path: Path,
    w_delays: WDelays | None = None,
) -> Path:
    with quantity_support(), time_support():
        before_amp_stokesi = np.abs(
            (
                before_baseline_data.masked_data[..., 0]
                + before_baseline_data.masked_data[..., -1]
            )
            / 2
        )
        after_amp_stokesi = np.abs(
            (
                after_baseline_data.masked_data[..., 0]
                + after_baseline_data.masked_data[..., -1]
            )
            / 2
        )

        norm = ImageNormalize(
            after_amp_stokesi, interval=ZScaleInterval(), stretch=SqrtStretch()
        )
        cmap = plt.cm.viridis

        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
            2, 2, figsize=(12, 10), sharex=True, sharey="row"
        )
        im = ax1.pcolormesh(
            before_baseline_data.time,
            before_baseline_data.freq_chan,
            before_amp_stokesi.T,
            norm=norm,
            cmap=cmap,
        )
        ax1.set(
            ylabel=f"Frequency / {before_baseline_data.freq_chan.unit:latex_inline}",
            title="Before",
        )
        ax2.pcolormesh(
            after_baseline_data.time,
            after_baseline_data.freq_chan,
            after_amp_stokesi.T,
            norm=norm,
            cmap=cmap,
        )
        ax2.set(
            ylabel=f"Frequency / {after_baseline_data.freq_chan.unit:latex_inline}",
            title="After",
        )
        for ax in (ax1, ax2):
            fig.colorbar(im, ax=ax, label="Stokes I Amplitude / Jy")

        # TODO: Move these delay calculations outside of the plotting function
        # And here we calculate the delay information

        before_delays_i = np.abs(
            (before_delays.delay_time[:, :, 0] + before_delays.delay_time[:, :, -1]) / 2
        )
        after_delays_i = np.abs(
            (after_delays.delay_time[:, :, 0] + after_delays.delay_time[:, :, -1]) / 2
        )

        delay_norm = ImageNormalize(
            before_delays_i, interval=MinMaxInterval(), stretch=LogStretch()
        )

        im = ax3.pcolormesh(
            before_baseline_data.time,
            before_delays.delay,
            before_delays_i.T,
            norm=delay_norm,
            cmap=cmap,
        )
        ax3.set(ylabel="Delay / s", title="Before")
        ax4.pcolormesh(
            after_baseline_data.time,
            after_delays.delay,
            after_delays_i.T,
            norm=delay_norm,
            cmap=cmap,
        )
        ax4.set(ylabel="Delay / s", title="After")
        for ax in (ax3, ax4):
            fig.colorbar(im, ax=ax, label="Stokes I Amplitude / Jy")

        if w_delays is not None:
            ant_1, ant_2 = before_baseline_data.ant_1, before_baseline_data.ant_2
            b_idx = w_delays.b_map[ant_1, ant_2]
            wrapped_w_delays = symmetric_domain_wrap(
                values=w_delays.w_delays[b_idx].value,
                upper_limit=np.max(after_delays.delay).value,
            )
            zones = calculate_nyquist_zone(
                values=w_delays.w_delays[b_idx].value,
                upper_limit=np.max(after_delays.delay).value,
            )
            # Final append is to capture the last zone in the
            # time range
            transitions = [*np.argwhere(np.diff(zones) != 0)[:, 0], len(zones)]

            for ax, baseline_data in zip(  # type:ignore[call-overload]
                (ax3, ax4),
                (before_baseline_data, after_baseline_data),
                strict=True,
            ):
                start_idx = 0
                for _zone_idx, end_idx in enumerate(transitions):
                    # The np.diff results in offset indices, so shift
                    # back the transition by 1
                    object_slice = slice(start_idx, end_idx + 1)
                    # and ensure non-overlapping line segments
                    start_idx = end_idx + 1
                    import matplotlib.patheffects as pe  # noqa: PLC0415

                    ax.plot(
                        baseline_data.time[object_slice],
                        wrapped_w_delays[object_slice],
                        color="tab:red",
                        # linestyle="-",
                        label=f"Delay for {w_delays.object_name}"
                        if _zone_idx == 0
                        else None,
                        dashes=(2 * _zone_idx + 1, 2 * _zone_idx + 1),
                        lw=4,
                        path_effects=[
                            pe.Stroke(
                                linewidth=6, foreground="white"
                            ),  # Add some contrast to help read line stand out
                            pe.Normal(),
                        ],
                    )
                ax.legend(loc="upper right")

        fig.suptitle(
            f"Ant {after_baseline_data.ant_1} - Ant {after_baseline_data.ant_2}"
        )
        fig.tight_layout()
        fig.savefig(output_path)

        return output_path



"""Helper utilities to deal with cyclic boundaries
on data"""

from __future__ import annotations

import numpy as np
from numpy.typing import NDArray


def symmetric_domain_wrap(
    values: NDArray[np.floating], upper_limit: float
) -> NDArray[np.floating]:
    """Place a set of values into a cyclic domain that is
    symmetric around zero.

    Args:
        values (NDArray[np.floating]): Values that need to be mapped into the cyclic domain
        upper_limit (float): The upper bound of the symmetric cyclic domain

    Returns:
        NDArray[np.floating]: Values that have been mapped to the -upper_limit to upper_limit domain
    """
    # Calculate an appropriate domain mapping
    # The natural domain is going to be mapped
    # to -pi to pi.
    domain_mapping = np.pi / upper_limit

    real = np.cos(values * domain_mapping)
    imag = np.sin(values * domain_mapping)

    wrapped_values = real + 1j * imag

    return np.angle(wrapped_values) / domain_mapping


def calculate_nyquist_zone(
    values: NDArray[np.floating], upper_limit: float
) -> NDArray[np.int_]:
    """Return the nyquist zone a value is in for a symmetric
    set of bounds around zero

    Args:
        values (NDArray[np.floating]): The values to calculate the zone for
        upper_limit (float): The upper bound to the symmetric domain around zero

    Returns:
        NDArray[np.int_]: The zones values correspond to
    """
    return np.array(
        np.floor((upper_limit + np.abs(values)) / (2.0 * upper_limit)) + 1, dtype=int
    )



"""Routines and structures to describe antennas, their
XYZ and baseline vectors"""

from __future__ import annotations

from argparse import ArgumentParser
from dataclasses import dataclass
from itertools import combinations
from pathlib import Path

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from casacore.tables import table

from jolly_roger.logging import logger


@dataclass
class Baselines:
    """Container representing the antennas found in some measurement set, their
    baselines and associated mappings. Only the upper triangle of
    baselines are formed, e.g. 1-2 not 2-1.
    """

    ant_xyz: np.ndarray
    """Antenna (X,Y,Z) coordinates taken from the measurement set"""
    b_xyz: np.ndarray
    """The baseline vectors formed from each antenna-pair"""
    b_idx: np.ndarray
    """Baselihe indices representing a pair of antenna"""
    b_map: dict[tuple[int, int], int]
    """A mapping between two antennas to their baseline index"""
    ms_path: Path
    """The measurement set used to construct some instance of `Baseline`"""


def get_baselines_from_ms(ms_path: Path) -> Baselines:
    """Extract the antenna positions from the nominated measurement
    set and constructed the set of baselines. These are drawn from
    the ANTENNA table in the measurement set.

    Args:
        ms_path (Path): The measurement set to extract baseliens from

    Returns:
        Baselines: The corresponding set of baselines formed.
    """

    logger.info(f"Creating baseline instance from {ms_path=}")
    with table(str(ms_path / "ANTENNA"), ack=False) as tab:
        ants_idx = np.arange(len(tab), dtype=int)
        b_idx = np.array(list(combinations(list(ants_idx), 2)))
        xyz = tab.getcol("POSITION")
        b_xyz = xyz[b_idx[:, 0]] - xyz[b_idx[:, 1]]

    b_map = {tuple(k): idx for idx, k in enumerate(b_idx)}

    logger.info(f"ants={len(ants_idx)}, baselines={b_idx.shape[0]}")
    return Baselines(
        ant_xyz=xyz * u.m, b_xyz=b_xyz * u.m, b_idx=b_idx, b_map=b_map, ms_path=ms_path
    )


@dataclass
class BaselinePlotPaths:
    """Names for plots for baseline visualisations"""

    antenna_path: Path
    """Output for the antenna XYZ plot"""
    baseline_path: Path
    """Output for the baselines vector plot"""


def make_plot_names(ms_path: Path) -> BaselinePlotPaths:
    """Construct the output paths of the diagnostic plots

    Args:
        ms_path (Path): The measurement set the plots are created for

    Returns:
        BaselinePlotPaths: The output paths for the plot names
    """

    basename = ms_path.parent / ms_path.stem

    antenna_path = Path(f"{basename!s}-antenna.pdf")
    baseline_path = Path(f"{basename!s}-baseline.pdf")

    return BaselinePlotPaths(antenna_path=antenna_path, baseline_path=baseline_path)


def plot_baselines(baselines: Baselines) -> BaselinePlotPaths:
    """Create basic diagnostic plots for a set of baselines. This
    includes the antenna positions and the baseline vectors.

    Args:
        baselines (Baselines): The loaded instance of the baselines from a measurement set

    Returns:
        BaselinePlotPaths: The output paths of the plots created
    """

    plot_names = make_plot_names(ms_path=baselines.ms_path)

    # Make the initial antenna plot
    fig, ax = plt.subplots(1, 1)

    ax.scatter(baselines.b_xyz[:, 0], baselines.b_xyz[:, 1], label="Baseline")

    ax.set(xlabel="X (meters)", ylabel="Y (meters)", title="ASKAP Baseline Vectors")
    ax.legend()

    fig.tight_layout()
    fig.savefig(plot_names.baseline_path)

    # Now plot the antennas
    fig, ax = plt.subplots(1, 1)

    ax.scatter(baselines.ant_xyz[:, 0], baselines.ant_xyz[:, 1], label="Antenna")

    ax.set(xlabel="X (meters)", ylabel="Y (meters)", title="ASKAP Antenna positions")
    ax.legend()
    fig.tight_layout()
    fig.savefig(plot_names.antenna_path)

    return plot_names


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Extract and plot antenna dna baseline information from a measurement set"
    )

    sub_parsers = parser.add_subparsers(dest="mode")

    plot_parser = sub_parsers.add_parser(
        "plot", description="Basic plots around baselines"
    )
    plot_parser.add_argument("ms_path", type=Path, help="Path to the measurement set")

    return parser


def cli() -> None:
    parser: ArgumentParser = get_parser()

    args = parser.parse_args()

    if args.mode == "plot":
        baselines = get_baselines_from_ms(ms_path=args.ms_path)
        plot_baselines(baselines=baselines)


if __name__ == "__main__":
    cli()



"""Utilities and structures around the delay calculations"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import astropy.units as u
import numpy as np
from numpy.typing import NDArray

from jolly_roger.logging import logger

if TYPE_CHECKING:
    # avoid circular imports
    from jolly_roger.tractor import BaselineData, DataChunk


@dataclass
class DelayTime:
    """Container for delay time and associated metadata."""

    delay_time: NDArray[np.complexfloating]
    """ The delay vs time data. shape=(time, delay, pol)"""
    delay: u.Quantity
    """The delay values corresponding to the delay time data."""


def data_to_delay_time(data: BaselineData | DataChunk) -> DelayTime:
    logger.debug("Converting freq-time to delay-time")
    delay_time = np.fft.fftshift(
        np.fft.fft(data.masked_data.filled(0 + 0j), axis=1), axes=1
    )
    delay = np.fft.fftshift(
        np.fft.fftfreq(
            n=len(data.freq_chan),
            d=np.diff(data.freq_chan).mean(),
        ).decompose()
    )
    return DelayTime(
        delay_time=delay_time,
        delay=delay,
    )


def delay_time_to_data(
    delay_time: DelayTime,
    original_data: DataChunk,
) -> DataChunk:
    """Convert delay time data back to the original data format."""
    logger.debug("Converting delay-time to freq-time")
    new_data = np.fft.ifft(
        np.fft.ifftshift(delay_time.delay_time, axes=1),
        axis=1,
    )
    new_data_masked = np.ma.masked_array(
        new_data,
        mask=original_data.masked_data.mask,
    )
    new_data = original_data
    new_data.masked_data = new_data_masked
    return new_data


@dataclass
class DelayRate:
    """Container for delay rate and associated metadata."""

    delay_rate: np.ndarray
    """The delay rate vs time data. shape=(rate, delay, pol)"""
    delay: u.Quantity
    """The delay values corresponding to the delay rate data."""
    rate: u.Quantity
    """The delay rate values corresponding to the delay rate data."""


def data_to_delay_rate(
    baseline_data: BaselineData,
) -> DelayRate:
    """Convert baseline data to delay rate."""
    # This only makes sense when running on time data. Hence
    # asserting the type of BaelineData

    assert isinstance(baseline_data, BaselineData), (
        f"baseline_data is type={type(baseline_data)}, but needs to be BaselineData"
    )

    logger.info("Converting freq-time to delay-rate")
    delay_rate = np.fft.fftshift(np.fft.fft2(baseline_data.masked_data.filled(0 + 0j)))
    delay = np.fft.fftshift(
        np.fft.fftfreq(
            n=len(baseline_data.freq_chan),
            d=np.diff(baseline_data.freq_chan).mean(),
        ).decompose()
    )
    rate = np.fft.fftshift(
        np.fft.fftfreq(
            n=len(baseline_data.time),
            d=np.diff(baseline_data.time.mjd * u.day).mean(),
        ).decompose()
    )

    return DelayRate(
        delay_rate=delay_rate,
        delay=delay,
        rate=rate,
    )
